# -*- coding: utf-8 -*-
# Copyright (c) 2023 Miroslav Hruska
"""
*********
Shortcuts
*********

Abbreviated names of common functions from :mod:`~daul.pandas_utils`.

Overview
********

Shortcuts: :class:`~pandas.DataFrame` columns
=============================================

.. autosummary::
   pduc
   pdutc
   pdufc
   pdca

Shortcuts: Conversion of :class:`~pandas.DataFrame`\ s to dictionaries
======================================================================

.. autosummary::
   pdtcd
   pdtcdf
   pdtcdfd

   pdtclvd
   pdtclvdf
   pdtclvdfd   

Shortcuts: Inner frames
=======================

.. autosummary::
   pdioc

Shortcuts: Other
================

.. autosummary::
   pdri

"""


#%% Modules
# global
from daul import pandas_utils as _pdu
  
#%% * Columns
pduc = _pdu.update_column
pdutc = _pdu.update_tuple_col
pdufc = _pdu.update_fixed_column 
pdca = _pdu.column_apply 

#%% * Dictionaries
pdtcd   = _pdu.twocol_dict 
pdtcdf  = _pdu.twocol_dictf
pdtcdfd = _pdu.twocol_dictfd

#%% * Listvaldicts
pdtclvd   = _pdu.twocol_listvaldict 
pdtclvdf  = _pdu.twocol_listvaldictf
pdtclvdfd = _pdu.twocol_listvaldictfd

#%% * Inner frames
pdioc = _pdu.include_outer_col

#%% * Other
pdri = _pdu.renumber_index 



# -*- coding: utf-8 -*-
# Copyright (c) 2023 Miroslav Hruska
"""
****************
Pandas utilities
****************

The module collects utilities for dealing with :class:`~pandas.DataFrame`\ s. 

Overview
********

Columns
=======

.. autosummary::
   rename_cols
   set_cols
   reorder_cols
   prefix_cols
   drop_cols
   safe_drop_cols

Column updates
==============

.. autosummary::
   update_column
   update_tuple_col
   update_fixed_column
   column_apply

Transformations
===============

.. autosummary::
   expand
   expand_dicts_as_cols

Inner frames
============

.. autosummary::
   groupby_as_frame
   extract_inner_col
   include_outer_col
   attach_inner_col
   attach_inner_cols
   rename_inner_cols

Row splitting
=============

.. autosummary::
   row_split_ngroups
   row_split_group_size
   row_split_sizes
   row_split_bool

Joins
=====

.. autosummary::
   left_join
   left_join_def

Conversion to dictionaries
==========================

.. autosummary::
   twocol_dict
   twocol_dictf
   twocol_dictfd
   twocol_listvaldict
   twocol_listvaldictf
   twocol_listvaldictfd

Selectors
=========

.. autosummary::
   nodup
   take_middle_n_rows
   ifst

Generic
=======

.. autosummary::
   empty_frame
   pdmap
   renumber_index
   values_list
   values_set

Empty-aware utilities
=====================

.. autosummary::
   eaw_row_concat

Compatibility
=============

.. autosummary::
   sort
   row_concat

"""

#%% Libraries
# global
import pandas as pd # skip
import numpy as np # skip
import math as _math
import functools as _ft

# local
import fplib as _fplib
from fplib import fst as _fst, snd as _snd

from daul import numpy_utils as _npu

#%% * Columns
def rename_cols(df, d):
  """
  Returns a copy of a frame `df` with 
  renamed columns as defined by the dictionary `d`. 
  
  Examples
  --------
  Let us create a an empty frame with two columns:

  >>> adf = pd.DataFrame({'a': [], 'b': []})

  Rename a copy of the frame:

  >>> bdf = rename_cols(adf, {'b': 'c'})
  >>> list(bdf.columns)
  ['a', 'c']
  
  Note that the column labels of the previous frame are unchanged: 
    
  >>> list(adf.columns)
  ['a', 'b']
  """
  # remove columns that are to be renamed to 
  # remove_cols = list(set(df.columns).intersection(d.values()))
  # df = drop_cols(df, remove_cols)  
  # set the column labels  
  df = set_cols(df, map(_fplib.dictfdf(d), df.columns))
  return df


def set_cols(df, cols):
  """
  Returns a copy of a frame `df` with its labels being `cols`. 
  
  Examples
  --------
  >>> adf = empty_frame(['a', 'b'])  
  >>> bdf = set_cols(adf, ['c', 'd'])
  >>> list(bdf.columns)
  ['c', 'd']
  
  Note that the columns of `adf` are not changed:
  
  >>> list(adf.columns)
  ['a', 'b']
  """
  ndf = df.copy()
  ndf.columns = cols
  return ndf


def reorder_cols(df, first_cols=[], last_cols=[]):
  """
  Returns a copy of frame `df` with reordered columns.

  The cols `first_cols` specifies the columns that will be the 
  put first, and `last_cols` the ones that will be the last. 
  
  Examples
  --------
  >>> adf = empty_frame(['a', 'b', 'c', 'd'])
  >>> bdf = reorder_cols(adf, ['b'], ['a'])
  >>> list(bdf.columns)
  ['b', 'c', 'd', 'a']
  
  Note that the order remains unchanged in the original frame.  
  
  >>> list(adf.columns)
  ['a', 'b', 'c', 'd']
  """
  cols = df.columns
  return df[first_cols 
            + filter(lambda col: not col in first_cols + last_cols, cols) 
            + last_cols]


def prefix_cols(df, pref):
  """
  Returns a copy of a frame `df` with prefix `pref` applied to column labels.

  Examples
  --------
  >>> df = pd.DataFrame({'a': [], 'b': []})
  >>> list(prefix_cols(df, 'l:').columns)
  ['l:a', 'l:b']
  """
  df = set_cols(df, map(lambda c: "%s%s" % (pref, c), df.columns))
  return df 


def drop_cols(df, cols):
  """
  Returns a copy of a frame `df` with columns `cols` dropped.
  
  Examples
  --------
  >>> df = empty_frame(['a', 'b', 'c'])  

  We can use it for one column:
  
  >>> list(drop_cols(df, 'a').columns)
  ['b', 'c']

  And also for multiple columns:  
  
  >>> list(drop_cols(df, ['a', 'b']).columns)
  ['c']
  """  
  return df.drop(cols, axis=1)


def safe_drop_cols(df, cols):
  """
  Returns a copy of a frame `df` with columns `cols` dropped, but 
  without raising exceptions if they do not exist.
  
  Examples
  --------
  >>> df = empty_frame(['a', 'b', 'c'])

  >>> list(safe_drop_cols(df, ['a', 'd']).columns)
  ['b', 'c']

  >>> list(safe_drop_cols(df, 'a').columns)
  ['b', 'c']
  """
  return df.drop(cols, axis=1, errors='ignore')


#%% * Column updates
def update_column(df, col, values, copy=True):
  """
  Returns a copy of a frame `df` with column `col` set to `values`.

  Parameters
  ----------
  copy: bool
    Whether to return a copy.
  
  Examples
  --------
  >>> adf = pd.DataFrame({'v': [0, 1, 2]})
  
  If a column with the desirable label does not exists, a new one is created:  
  
  >>> bdf = update_column(adf, 'e', ['a', 'b', 'c'])
  >>> list(bdf['e'])
  ['a', 'b', 'c']
  
  If it does, the column will have new values:
  
  >>> cdf = update_column(bdf, 'e', ['d', 'e', 'f'])
  >>> list(cdf['e'])
  ['d', 'e', 'f']
  
  Note that the previous frame remains the same.  
  
  >>> list(bdf['e'])
  ['a', 'b', 'c']
  """
  if copy:
    df = df.copy()
  df[col] = list(values)
  return df


def update_tuple_col(df, cola, colb, values, copy=True):
  """
  Returns a copy of a frame `df` and sets values of two columns
  (`cola`, `colb`) from a list of 2-len tuples (`values`).

  Parameters
  ----------
  copy: bool
    Whether to return a copy.
  
  Examples
  --------
  >>> df = pd.DataFrame({'id': [0, 1]})
  >>> values = [('green', '#00ff00'), ('blue', '#0000ff')]
  >>> df = update_tuple_col(df, 'name', 'hex', values)  
  >>> df
     id   name      hex
  0   0  green  #00ff00
  1   1   blue  #0000ff
  """
  if copy:
    df = df.copy()
  df[cola] = map(_fst, values)
  df[colb] = map(_snd, values)
  return df


def update_fixed_column(df, col, value, copy=True):
  """
  Returns a copy of a frame `df` with a column `col` set to a 
  fixed value `value`. 
  
  See also :func:`daul.shortcuts.pdufc` for a shorter form.  

  Parameters
  ----------
  copy: bool
    Whether to return a copy.
  
  Examples
  --------
  >>> df = pd.DataFrame({'a': [0, 1]})
  >>> df = update_fixed_column(df, 'b', 'text')
  >>> list(df['b'])
  ['text', 'text']
  
  .. warning:: 
    Be aware that the objects assigned are identical and 
    if mutable, changing one will result in the change of others.
    
  >>> df = pd.DataFrame({'a': [0, 1]})
  >>> df = update_fixed_column(df, 'b', [])
  >>> list(df['b'])
  [[], []]
  
  >>> df['b'].iloc[0].append(1) # change the first object
  >>> list(df['b'])
  [[1], [1]]
  """
  df = update_column(df, col, 
                     map(_fplib.constf(value), np.arange(len(df))),
                     copy=copy)
  return df


def column_apply(df, col, f, store_as=None, copy=True):
  """
  Returns a copy of a frame with a function `f` being applied over a column
  `col` and stored as the same column `col`. 
  
  Parameters
  ----------
  df : :class:`~pandas.DataFrame`
    The frame. 

  col : 
    The column over which to apply `f`. 
    
  f : function
    The function to apply over `col`.
    
  store_as : column-label or None
    The label of the column to store results of applying 
    `f`. If None, it is the same as `col`.
    
  copy : bool
    Whether to return a copy of the frame. 
  
  Examples
  --------
  >>> df = pd.DataFrame({'a': [0, 1]})
  >>> df = column_apply(df, 'a', lambda x: x + 1)
  >>> df
     a
  0  1
  1  2
  """
  store_col = col if store_as is None else store_as
  df = update_column(df, store_col, df[col].apply(f), copy=copy)
  return df


#%% * Transformations
def expand(df, col, exp_col=None, remove_col=False):
  """
  Expands the values of a list-like column into separate rows,
  while keeping the rest of the columns fixed.
  
  Parameters
  ----------
  df : :class:`~pandas.DataFrame`
    The frame to expand.   
  
  col : str
    A column of `df` to expand.
    
  exp_col : str
    The label to store the expanded column as.
    If None, then name of `col` used.

  remove_col : bool
    Indicates whether to remove the non-expanded column.
  
  Examples
  --------
  >>> df = pd.DataFrame({'a': [1, 2], 
  ...                    'fruits': [['orange', 'apple'], 
  ...                               ['kiwi']]})
  >>> df
     a           fruits
  0  1  [orange, apple]
  1  2           [kiwi]
  >>> df = expand(df, 'fruits', 'fruit')
  >>> df
     a           fruits   fruit
  0  1  [orange, apple]  orange
  0  1  [orange, apple]   apple
  1  2           [kiwi]    kiwi
  >>> 
  
  """
  I = np.repeat(np.arange(len(df)), 
                np.array(df[col].apply(len).values, dtype=np.int))
  edf = df.iloc[I]
  if exp_col is None:
    exp_col = col
  edf = update_column(edf, exp_col, _fplib.unlist1(df[col]))
  if remove_col:
    edf = edf.drop([col], axis=1)
  return edf


def expand_dicts_as_cols(df, col, remove_col=False):
  """
  Returns a frame in which the values of a dict-based column `col` in 
  frame `df` are expanded into separate columns.
  
  Parameters
  ----------
  remove_col : bool
    Whether to remove the dict-based column.   
  
  Examples
  --------
  Suppose a frame with dict-like column `v`.
  
  >>> df = pd.DataFrame({'i': [0, 1], 
  ...                    'v': [{'a': 1, 'b': 2}, 
  ...                          {'a': 2, 'b': 3}]})
  >>> df
     i                   v
  0  0  {u'a': 1, u'b': 2}
  1  1  {u'a': 2, u'b': 3}

  Now let us expand the dictionaries in `v`. 

  >>> df = expand_dicts_as_cols(df, 'v', remove_col=True)
  >>> df[['i', 'a', 'b']]
     i  a  b
  0  0  1  2
  1  1  2  3
  """
  ndf = pd.DataFrame(list(df[col]))
  for ecol in ndf.columns:
    df = update_column(df, ecol, list(ndf[ecol]))
  if remove_col:
    df = df.drop([col], axis=1)
  return df


#%% * Inner frames
def groupby_as_frame(df, col, df_col='df'):
  """
  Groups a frame by a column and stores the frames as inner frames. 
  
  Parameters
  ----------
  df : :class:`~pandas.DataFrame`
    The frame to group. 
    
  col : label
    The column by which to group. 
    
  df_col : label
    The label that will hold the resulting subframes. 

  Examples
  --------
  >>> df = pd.DataFrame({'a': [0, 0, 1], 'b': ['a', 'b', 'c']})
  >>> df
     a  b
  0  0  a
  1  0  b
  2  1  c
  
  Let us group the frame on the `a` column:

  >>> gdf = groupby_as_frame(df, 'a')
  
  Let us see the groups: 
  
  >>> gdf[['a']]
     a
  0  0
  1  1
  
  The first group (with `a = 0`):  
  
  >>> gdf['df'].iloc[0]
     a  b
  0  0  a
  1  0  b

  The second group (with `a = 1`):

  >>> gdf['df'].iloc[1]
     a  b
  2  1  c
  """  
  if len(df) == 0:
    return pd.DataFrame({df_col: [], col: []})
  gdf = pd.DataFrame(list(df.groupby(col)))
  gdf = set_cols(gdf, [col, df_col])
  return gdf


def extract_inner_col(df, df_col, inndf_col, aggf=list):
  """
  Extracts values from an inner frame's column 
  and applies a function over them. 
  
  Parameters
  ----------
  df : :class:`~pandas.DataFrame`
    The frame.

  df_col : str
    The label of column in `df` that holds the inner frames. 
    
  inndf_col : str
    The label of column in the inner frames. 
    
  aggf : function
    The function to aggregate the columns. 
  
  Examples
  --------
  Let us first create a frame and then group it (see :func:`groupby_as_frame`), 
  such that is has an inner frame:
  
  >>> df = pd.DataFrame({'a': [0, 0, 1, 2], 'b': [1, 3, 5, 7]})
  >>> gdf = groupby_as_frame(df, 'a') # inner frame in `df`

  Now let us extract the values from the columns `b` from the inner frame:

  >>> extract_inner_col(gdf, 'df', 'b')
  0    [1, 3]
  1       [5]
  2       [7]
  Name: df, dtype: object
  
  We can also apply some function, e.g., to sum the values:  
  
  >>> extract_inner_col(gdf, 'df', 'b', sum)
  0    4
  1    5
  2    7
  Name: df, dtype: int64
  """
  return df[df_col].apply(lambda df: aggf(df[inndf_col]))


def include_outer_col(df, inndf_col, col):
  """
  Includes column from the outer frame as a fixed column in the inner frame.

  Parameters
  ----------
  df : :class:`~pandas.DataFrame`
    Frame to update. 
    
  inndf_col : label
    The label of the column that holds the inner frames. 
    
  col : label
    The label of the column which to include from `df` to the inner frames. 
  """
  df = update_column(
      df, 
      inndf_col, 
      map(lambda col_value, inndf: 
        update_fixed_column(inndf, col, col_value), df[col], df[inndf_col]))
  return df


def attach_inner_col(df, col, df_col, inndf_col, aggf=list):
  """
  Attaches a column to the outer frame by 
  aggregating the values in the inner frame.
  
  Parameters
  ----------
  df : :class:`~pandas.DataFrame`
    The frame to attach the column to.

  col : str  
    The label of the column that will be attached to the outer frame.
    
  df_col : str
    The label of the column where the inner frames are stored.

  inndf_col : str
    The label of the column in the inner frame, which will be aggregated. 
    
  aggf : function
    The function that aggregates the values. 
  """
  df = update_column(df, col, 
                     df[df_col].apply(lambda inndf: aggf(inndf[inndf_col])))
  return df


def attach_inner_cols(df, cols, df_col, aggf, namef=_fplib.idf):
  """
  Attaches selected columns from the inner frame. 
  
  Parameters
  ----------
    df : :class:`~pandas.DataFrame`
      The frame to attach the columns to. 
      
    cols : list
      The labels of the columns in the inner frame to attach to the 
      outer frame. See also `namef` parameter.
      
    df_col : str
      The label of the column where the inner frames are stored.
      
    aggf : function
      The function that aggregates the values from the inner frames.
      
    namef : function
      The function that maps labels of the columns from the inner frame
      to the labels in the outer frame. 
  
  Examples
  --------
  Let us create a simple frame that we will group to obtain inner frames:
  
  >>> df = pd.DataFrame({'a': [0, 0, 1, 2], 'b': ['a', 'b', 'c', 'd']})
  >>> df
     a  b
  0  0  a
  1  0  b
  2  1  c
  3  2  d

  Group it: 

  >>> gdf = groupby_as_frame(df, 'a') # inner frame is in 'df' column
  
  Now we can attach the inner columns as outer ones:
  
  >>> gdf = attach_inner_cols(gdf, ['b'], 'df', list)
  >>> gdf[['a', 'b']]
     a       b
  0  0  [a, b]
  1  1     [c]
  2  2     [d]
  """
  for col in cols:
    df = attach_inner_col(df, namef(col), df_col, col, aggf)
  return df


def rename_inner_cols(df, df_col, d={}):
  """
  Renames columns of inner frames.
  
  Parameters
  ----------
  df : :class:`~pandas.DataFrame`
    The frame to transform.
    
  df_col: str
    The label of the column where the inner frames are stored.
    
  d : dict
    The renaming dictionary.
  """
  df = column_apply(df, df_col, _ft.partial(rename_cols, d=d))
  return df


#%% * Row spliting
def _row_split_empty(df, empty):
  if not (empty in ['nogroup', 'zerolengroup']):
    raise ValueError("The *empty* must be one of nogroup or zerolengroup.")
  if empty == 'zerolengroup':
    return [df]
  elif empty == 'nogroup':
    return []


def row_split_ngroups(df, n, empty='zerolengroup'):
  """
  Row-splits the frame into `n` approximately equal-length frames. 

  Parameters
  ----------
  df : :class:`~pandas.DataFrame`
    The frame to row-split.
    
  n : int
    The number of groups.

  empty : str
    Applies only if the frame to split is of zero length. 
    Using 'zerolengroup' returns one group of zero length 
    (with the column labels preserved). 
    Using 'nogroup' returns an empty list.
  
  .. note:: 
    Uses NumPy :func:`array_split` for splitting the frame. 

  .. note::
    If the number of rows `r` in the frame is less than `n`, returns
    `r` groups. 

  Examples
  --------
  >>> df = pd.DataFrame({'x': np.arange(0, 1000)})
  >>> n = 10  
  >>> dfs = row_split_ngroups(df, n)
  
  The total number of frames:
  
  >>> len(dfs)
  10
  
  The size of the first frame:
  
  >>> len(dfs[0])
  100

  Note the behavior if the number of groups is less than rows:
  
  >>> df = pd.DataFrame({'x': np.arange(0, 5)})
  >>> dfs = row_split_ngroups(df, 10)
  >>> len(dfs)
  5
  
  In case of an empty frame, the result is a one group of 
  with an empty frame, by default: 
  
  >>> df = empty_frame(['a', 'b'])
  >>> dfs = row_split_ngroups(df, 5, empty='zerolengroup')
  >>> len(dfs)
  1
  >>> len(dfs[0])
  0
  
  In this case the shape of the frame is preserved and thus 
  further processing of the frame will likely succeed:

  >>> list(dfs[0].columns)
  ['a', 'b']
  
  In case of 'nogroup', returns an empty list:
  
  >>> row_split_ngroups(df, 5, empty='nogroup')  
  []
  """
  if len(df) == 0:
    return _row_split_empty(df, empty=empty)
  locs = np.array_split(np.arange(len(df), dtype=np.int), n)
  locs = filter(lambda x: len(x) > 0, locs)
  dfs  = map(lambda iloc: df.iloc[iloc], locs)
  return dfs


def row_split_group_size(df, sz, empty='zerolengroup'):
  """
  Row-splits the frame into frames having `sz` rows. 
  
  Parameters
  ----------
  df : :class:`~pandas.DataFrame`
    The frame to row-split. 
    
  sz : int
    The number of rows in a group.
  
  empty : str
    Applies only if the frame to split is of zero length. 
    Using 'zerolengroup' returns one group of zero length 
    (with the column labels preserved). 
    Using 'nogroup' returns an empty list.
    
  Examples
  --------
  >>> df = pd.DataFrame({'x': np.arange(0, 1000)})
  >>> dfs = row_split_group_size(df, 5)  
  >>> len(dfs)
  200
  >>> len(dfs[0])
  5

  .. seealso:: 
     :func:`row_split_ngroups` for examples when the frame to split 
     has zero rows.
  """
  if len(df) == 0:
    return _row_split_empty(df, empty=empty)
  locs = np.array_split(np.arange(len(df), dtype=np.int), 
                        _math.ceil(1.0 * len(df) / sz))
  dfs = map(lambda iloc: df.iloc[iloc], locs)
  return dfs


def row_split_sizes(df, lens):
  """
  Splits a frame `df` according to a list `lens` of lengths.
  
  Examples
  --------
  >>> df = pd.DataFrame({'x': np.arange(0, 10)})  
  >>> lens = [1, 3, 2, 4]
  
  Split and check whether the lengths correspond. 
  
  >>> dfs = row_split_sizes(df, lens)
  >>> map(len, dfs)
  [1, 3, 2, 4]

  Let us take a look at the last frame:

  >>> dfs[-1]
     x
  6  6
  7  7
  8  8
  9  9
  """
  if sum(lens) != len(df):  
    raise ValueError("The sum of the lens must be the same as " \
                     "the len of the frame.")
  nf = _npu.zero_cumsum_nolast(lens)
  nt = np.cumsum(lens)
  return map(lambda s, e: df[s:e], nf, nt)


def row_split_bool(df, arr):
  """
  Row-splits a frame `df` into two parts: 
  the rows for which the `arr` is True and those for which it is False.
  
  Returns
  -------
   tdf, fdf : tuple  
  
  Examples
  --------  
  >>> df = pd.DataFrame({'a': np.arange(0, 10)})
  
  Split the frame into those with an even and odd numbers:
  
  >>> adf, bdf = row_split_bool(df, df['a'] % 2 == 0)
  >>> adf  
     a
  0  0
  2  2
  4  4
  6  6
  8  8
  """
  if len(df) != len(arr):  
    raise ValueError("The length of the boolean array must be the same " \
                     "as that of the frame.")
  arr = np.array(arr, dtype=np.bool)
  tdf = df.loc[ arr]
  fdf = df.loc[~arr]
  return tdf, fdf


#%% * Joins
def left_join(ldf, rdf, jcol, rcols=[]):
  """
  Returns a copy of the left frame with columns joined from the right frame on
  a specified column.
  
  Parameters
  ----------
    ldf : :class:`~pandas.DataFrame`
      The left frame.
      
    rdf : :class:`~pandas.DataFrame`
      The right frame. 
      
    jcol : str
      The label of the column used on which to perform the join.
      
    rcols: list
      The columns from `rdf` to join. 
      

  .. note::
    No default behavior if the particular value in `jcol` is missing 
    in the right frame; see :func:`left_join_def`.

  Examples
  --------
  >>> ldf = pd.DataFrame({'a': [0, 0, 1, 2]})
  >>> rdf = pd.DataFrame({'a': [0, 1, 2], 'b': ['zero', 'one', 'two']})
  >>> left_join(ldf, rdf, 'a', ['b'])
     a     b
  0  0  zero
  1  0  zero
  2  1   one
  3  2   two
  """
  for rcol in rcols:
    ldf = update_column(ldf, rcol, 
                        ldf[jcol].apply(twocol_dictf(rdf[[jcol, rcol]])))
  return ldf


def left_join_def(df, jdf, jcol, cols=[], default=None):
  """
  Returns a copy of the left frame with columns joined from the right frame,
  providing a default value. 
  
  .. note::
    See :func:`left_join` for explanation of parameters.
  """  
  for col in cols:
    df = update_column(df, col, 
                       df[jcol].apply(twocol_dictfd(jdf[[jcol, col]], 
                                                    default=default)))
  return df


#%% * Conversion to dictionaries
def twocol_dict(df):
  """
  Creates a dictionary from the first two columns of a frame `df` 
  (`keys` --- the first column, `values` --- the second column). 
  
  .. seealso:: 
    :func:`twocol_listvaldict` for creating a list-valued dictionary. 
  
  Examples
  --------
  >>> df = pd.DataFrame({'k': ['a', 'b'], 'v': ['x', 'y']})[['k', 'v']]
  >>> df
     k  v
  0  a  x
  1  b  y
  >>> twocol_dict(df)
  {'a': 'x', 'b': 'y'}
  """
  return dict(zip(df[_fst(df.columns)], df[_snd(df.columns)]))


def twocol_dictf(df):
  """
  Creates a function that maps values from the first column of `df` to the 
  values in the corresponding rows of the second column.
  
  .. note::
    Creates a dictionary using :func:`twocol_dict`. 
  """  
  return _fplib.dictf(twocol_dict(df))


def twocol_dictfd(df, default):
  """
  Creates a function that maps values from the first column of `df` to the 
  values in the corresponding rows of the second column, 
  with a default value `default`. 
  """
  return _fplib.dictfd(twocol_dict(df), default=default)

# List-valued dictionaries

def twocol_listvaldict(df):
  """
  Creates a dictionary from the first two columns of a frame `df`, 
  (`keys` --- the first column, `values` --- list-aggregated values
  from the second column for the same `key`).
  
  .. seealso:: :func:`twocol_dict`. 
  
  Examples
  --------
  >>> df = pd.DataFrame({'k': ['a', 'a', 'b'], 'v': ['c', 'd', 'e']})[['k', 'v']]
  >>> df
     k  v
  0  a  c
  1  a  d
  2  b  e
  >>> twocol_listvaldict(df)
  {'a': ['c', 'd'], 'b': ['e']}
  """
  return _fplib.listvaldict(zip(df[_fst(df.columns)], df[_snd(df.columns)]))
  

def twocol_listvaldictf(df):
  """
  Creates a function that maps `keys` from the first column of `df` 
  into a `list` of `values` from the second column corresponding to the same `key`.
  
  .. seealso::
    The function is a wrapper over :func:`twocol_listvaldict`.
  """  
  return _fplib.dictf(twocol_listvaldict(df))
  
  
def twocol_listvaldictfd(df, default):
  """
  Creates a function that maps `keys` from the first column of `df` into a 
  `list` of `values` from the second column corresponding to the same `key`
  (with a default value `default`, if no such `key` is found).

  .. seealso::
    The function is a wrapper over :func:`twocol_listvaldict`.
  """
  return _fplib.dictfd(twocol_listvaldict(df), default=default)


#%% * Selectors
def take_middle_n_rows(df, n, error_if_less=False):
  """
  Selects `n` middle rows of a frame.

  :raises: ValueError
    
    If the `df` has less than `n` rows, and if 
    `error_if_less` is set. 

  Examples
  --------
  >>> df = pd.DataFrame({'col': ['a', 'b', 'c']})  
  >>> take_middle_n_rows(df, 1)
    col
  1   b
  >>> take_middle_n_rows(df, 2)
    col
  0   a
  1   b
  """  
  if len(df) >= n:
    return df.iloc[((len(df) - n) / 2): ((len(df) + n) / 2)]
  else:
    if error_if_less:
      raise ValueError("Not enough rows.")
    else:
      return df


def ifst(x):
  """
  Returns the first element of `x` using `iloc` of a frame, or a series.
  """
  return x.iloc[0]


def nodup(df, cols=None):
  """
  Returns non-duplicated rows of a frame `df`.
  
  If the `cols` is :code:`None`, then all columns are used.  
  """
  if cols is None:
    cols = df.columns
  return df.loc[-df[cols].duplicated()]


#%% * Empty-aware utilities
def eaw_row_concat(dfs, cols):
  """ 
  Performs a row-wise concatenation of frames `dfs`. If an empty list is given, 
  returns an empty frame with the specified columns `cols`.
  """
  if sum(map(len, dfs)) == 0:
    return empty_frame(cols)
  else:
    return row_concat(dfs)



#%% Generic
def empty_frame(cols=[]):
  """
  Creates an empty frame with the given column labels `cols`. 
  
  Examples
  --------
  >>> df = empty_frame(['a', 'b'])  
  >>> len(df)
  0
  >>> list(df.columns)
  ['a', 'b']
  """
  return pd.DataFrame(dict(_fplib.zipe(cols, [])))


def renumber_index(df, copy=True):
  """
  Returns a copy of a frame `df`, with the index having consecutive values in 
  the range from :code:`0` to :code:`len(df) - 1`. 
  
  Parameters
  ----------
  copy : bool
    Whether to return a copy of the frame. 

  """
  if copy:  
    df = df.copy()
  df.index = np.arange(len(df))
  return df


def pdmap(f, df):
  """
  Maps a function `f` over rows of a frame `df`, with the function application 
  being done over all columns as positional arguments. 
  
  Examples
  --------
  >>> df = pd.DataFrame({'a': [0, 1]})
  >>> df = update_column(df, 'b', df['a'] + 10)
  >>> df
     a   b
  0  0  10
  1  1  11
  >>> pdmap(lambda x, y: x + y, df)
  [10, 12]
  """
  cols = map(lambda i: df[df.columns[i]].values, range(len(df.columns)))
  return map(f, *cols)
  

def values_list(x):
  """
  Returns the `x.values` of `x` as a list.
  
  Examples
  --------
  >>> df = pd.DataFrame({'x': np.arange(5)})  
  >>> values_list(df['x'])
  [0, 1, 2, 3, 4]
  """
  return list(x.values)


def values_set(x):
  """
  Returns the `x.values` of `x` as a set.
  
  Examples
  --------
  >>> df = pd.DataFrame({'x': ['a', 'a', 'a']})  
  >>> values_set(df['x'])
  set(['a'])
  """
  return set(x.values)
  
  
#%% * Compatibility
def sort(df, col, ascending=True):
  """
  Sorts frame by given col(s).
  
  Parameters
  ----------
  df : :class:`~pandas.DataFrame`
    The frame to sort. 
    
  col : str or list
    Columns by which to sort. 
  
  ascending : bool or list
    Whether to sort in ascending order. 
  
  .. note:: 
    Internally uses :func:`pd.DataFrame.sort_values` 
    or :func:`pd.DataFrame.sort` if not available.
  """
  if hasattr(df, 'sort_values'):
    return df.sort_values(col, ascending=ascending)
  else:
    return df.sort(col, ascending=ascending)
    

def row_concat(dfs):
  """
  Concatenates frames `dfs`, row-wise.
  """
  if not 'sort' in pd.concat.func_code.co_varnames:
    return pd.concat(list(dfs))
  else:
    return pd.concat(list(dfs), sort=False)


#%% * Empty-aware
def eaw_groupby_agg(df, groupby, aggd):
  """
  Groups a frame and aggregates values, even for an empty frame. 

  Parameters
  ----------
  df : :class:`~pandas.DataFrame`
    Frame to group. 
    
  groupby : str
    Label of column to groupby. 
    
  aggd : dict
    Dictionary of labels and functions to as in 
    :func:`pandas.DataFrame.agg`. 

  Examples
  --------
  >>> df = pd.DataFrame({'x': [0, 0, 1], 'y': [1, 2, 3]})
  >>> df
     x  y
  0  0  1
  1  0  2
  2  1  3
  >>> gdf = eaw_groupby_agg(df, 'x', {'y': np.max}).reset_index()
  >>> gdf
     x  y
  0  0  2
  1  1  3
  >>> df = empty_frame(['x', 'y'])
  >>> gdf = eaw_groupby_agg(df, 'x', {'y': np.max}).reset_index()
  >>> gdf
  Empty DataFrame
  Columns: [x, y]
  Index: []
  """
  if len(df) == 0:
    gdf = pd.DataFrame(dict(_fplib.zipe(aggd.keys(), [])), index=pd.Index([], name=groupby))
    return gdf
  else:
    return df.groupby(groupby).agg(aggd)


#%% Main
if __name__ == '__main__':
  import doctest
  doctest.testmod()



# -*- coding: utf-8 -*-
# Copyright (c) 2023 Miroslav Hruska
"""
**********
File utils
**********

The module provides convenience functions for dealing with files, 
directories and paths.

Overview
********

Reading
=======

.. autosummary::
   read_file_lines
   read_file_first_line
   read_pickle_file

Writing
=======   
   
.. autosummary::
   write_lines_file
   write_pickle_file
   
File extensions
===============

.. autosummary::
   file_ext
   set_ext
   file_complete_ext
   set_complete_ext
   remove_complete_ext
   
Directory listing
=================

.. autosummary::
   list_dir
   walk_dir

Paths
=====

.. autosummary::
   abs_path
   prepend_dir

Directories
===========

.. autosummary::
   safe_create_dir

Sizes
=====

.. autosummary::
   file_size
   dir_size

Symlinks
========

.. autosummary::
   rel_symlink

Utilities
=========

.. autosummary::
   file_frame
   random_hash

"""

#%% Modules
# global
import os # skip
import pathlib # skip
import pickle as _pickle
import pandas as _pd
import functools as _ft
import binascii as _binascii
import fnmatch as _fnmatch

# local
import fplib as _fplib
from daul import pandas_utils as _pdu

#%% * Reading
def read_file_lines(f, strip=False, remove_empty=False, nlines=None):
  """
  Reads all lines of a given file.
  
  Parameters
  ----------
  f : str
    Path to the file.   
    
  strip: bool
    Indicates whether to strip lines of white space.
    
  remove_empty: bool
    Indicates whether to remove empty lines.
    
  nlines : int or None
    If ``None``, reads all the lines. 
    If ``int``, reads the specified number of lines.
  """
  with open(f, "r") as h:
    if nlines is None:    
      lns = h.readlines()
    else:
      lns = []
      for i in range(nlines):
        lns.append(h.readline())
    if strip:
      lns = map(lambda ln: ln.strip(), lns)
    if remove_empty:
      lns = filter(lambda ln: len(ln) > 0, lns)
    return lns
    

def read_file_first_line(f):
  """
  Reads the first line of a given file `f`.
  """
  with open(f, "r") as h:
    return h.readline()
    

def read_pickle_file(f):
  """
  Reads a pickled object from a file `f`.
  """  
  with open(f, "r") as h:
    return _pickle.load(h)
    
    
#%% * Writing
def write_lines_file(lns, f, linedel=''):
  """
  Writes lines into a file.
  
  Parameters
  ----------
  lns : list
    List of lines. 
  
  f : str
    Path to the file. 
    
  linedel : str
    Line delimiter. 
  """
  with open(f, "w") as h:
    if linedel != '':
      lns = map(lambda ln: ln + linedel, lns)
    h.writelines(lns)

  
def write_pickle_file(e, f):
  """
  Writes an object `e` as a pickle into a file `f`.
  """
  with open(f, "w") as h:
    _pickle.dump(e, h)
  

#%% * File extensions
def file_ext(f):
  """
  Returns the extension of a file `f`.
  
  Examples
  --------
  >>> file_ext('abc.txt.gz')
  '.gz'
  >>> file_ext('abc')
  ''
  """    
  return _fplib.snd(os.path.splitext(f))


def set_ext(f, ext=''):
  """
  Changes the file extension of path `f` to `ext` by replacing the 
  text starting at the last period.

  Examples
  --------
  >>> set_ext('abc.txt', '.tex')
  'abc.tex'
  >>> set_ext('abc', '.txt')
  'abc.txt'
  >>> set_ext('/home/user/file', '.cache')
  '/home/user/file.cache'

  The changes are applied only to the basename: 

  >>> set_ext('/home/user/.conf/file', '.cache')
  '/home/user/.conf/file.cache'
  
  If multiple periods are present in the basename, 
  it changes only the last one:
  
  >>> set_ext('file.txt.cache', '.ch')
  'file.txt.ch'
  
  .. seealso:: :func:`set_complete_ext` for changing the complete extension. 
  """
  basename, prev_ext = os.path.splitext(f)
  return basename + ext


def file_complete_ext(f):
  """
  Returns the complete extension of a file `f`.
  
  Examples
  --------
  >>> file_complete_ext('abc.txt.gz')
  '.txt.gz'
  """
  bn = os.path.basename(f)
  p = bn.find('.')
  if p == -1:
    return ''
  else:
    return bn[p:]
    
    
def set_complete_ext(f, ext=''):
  """
  Changes the file extension of path `f` by changing the text after 
  the first period in its basename to `ext`.
  
  Examples
  --------
  >>> set_complete_ext('abc.txt', '.tex')
  'abc.tex'
  >>> set_complete_ext('abc.txt.gz', '.ext')
  'abc.ext'
  >>> set_complete_ext('abc.txt.gz', '')
  'abc'
  
  .. seealso:: :func:`remove_complete_ext` for removing the complete extension.
  """
  dn = os.path.dirname(f)
  bn = os.path.basename(f)
  p = bn.find('.')
  if p == -1:
    return "%s%s" % (os.path.join(dn, bn), ext)
  else:
    return os.path.join(dn, bn[:p] + ext)
  

def remove_complete_ext(f):
  """
  Removes the complete extension of a path `f`.
  
  Examples
  --------  
  >>> remove_complete_ext('abc.txt.gz')
  'abc'
  """
  return set_complete_ext(f, "")


#%% * Directory listings
def list_dir(d, match='*', sort=False):
  """
  Lists files and directories that match a given pattern.
  
  The pattern is checked against the basename of the file/dir.    
  
  Parameters
  ----------
  d : str
    Directory to list. 
    
  match : str
    Name pattern that the files/dirs must match to be included.
    
  sort : bool
    Indicates whether to sort files/dirs by name.
  """
  fls = filter(lambda x: _fnmatch.fnmatch(os.path.basename(x), match), 
               map(_ft.partial(os.path.join, d), os.listdir(d))) 
  if sort:
    fls = sorted(fls)
  return fls


def walk_dir(d, match='*', follow_symlinks=False):
  """
  Walks a given directory and returns files/dirs whose basename 
  matches a given pattern.
  
  Parameters
  ----------
  d : str
    Path to the directory to walk.
    
  match : str
    The pattern of files to return. 
    
  follow_symlinks : bool
    Whether to follow links as in :func:`os.walk`.
  """
  recs = list(os.walk(d, followlinks=follow_symlinks))
  pathss = map(_fplib.npify(lambda base, dirs, files: 
              map(lambda path: os.path.join(base, path), dirs + files)), 
              recs)
  paths = sorted([d] + _fplib.unlist1(pathss))
  paths = filter(lambda x: _fnmatch.fnmatch(os.path.basename(x), match), 
                 paths)
  return paths


#%% * Paths
def as_posix_path(path):
  """
  Returns the path `path` as a POSIX path.
  
  Examples
  --------
  >>> as_posix_path('C:\\Windows')
  'C:/Windows'
  >>> as_posix_path('/home/user')
  '/home/user'
  >>> as_posix_path('C:\\Windows/system32')
  'C:/Windows/system32'
  """
  if '\\' in path:
    return pathlib.PureWindowsPath(path).as_posix()
  else:
    return pathlib.PurePosixPath(path).as_posix()


def abs_path(path, d):
  """
  Returns an absolute POSIX path of `path` given a working directory `d`. 
  
  Examples
  --------  
  >>> abs_path('dir/file', '/base')
  '/base/dir/file'
  """
  if not os.path.isabs(path):
    path = os.path.join(d, path)
  return as_posix_path(path)


def prepend_dir(path, d):
  """
  Prepends a directory `d` just before the path `path`.
  
  Examples
  --------
  >>> prepend_dir('file', 'dir')
  'dir/file'
  >>> prepend_dir('path/file', 'dir')
  'path/dir/file'
  >>> prepend_dir('/home/user/file', 'dir')
  '/home/user/dir/file'
  """
  path = os.path.join(os.path.dirname(path), d, os.path.basename(path))
  return as_posix_path(path)


#%% * Directories
def safe_create_dir(d):
  """
  Creates a directory, if it does not already exist.
  
  Parameters
  ----------
  d : str
    Path to the directory to create.


  :raises OSError: 
    If the `d` exists and it is not a directory, 
    throws an exception. 
  """
  if os.path.exists(d) and os.path.isdir(d):
    return
  os.makedirs(d)
  

#%% * Sizes
def file_size(f):
  """
  Returns the size of a file `f`, in bytes.
  """
  return os.stat(f).st_size


def dir_size(d, recursive=False):
  """
  Returns the sum of size of files in a directory `d`, in bytes.
  
  Parameters
  ----------
  recursive : bool
    Whether to recurse into subdirectories. 
  """
  fn = list_dir if not recursive else walk_dir
  return sum(map(file_size, fn(d)))
  

#%% Utilities * 
def file_frame(fs, bf=True, bfne=False, size=False, stat=False):
  """ 
  Constructs a :class:`~pandas.DataFrame` from file names, having selected columns.

  Parameters
  ----------
  fs : list of str
    a list of file names.  
  
  bf : bool
    whether to include the basename of a file (as "bf" column).
  
  bfne : bool
    whether to include the basename of a file without complete extension 
    (as "bfne" column).
    
  size : bool
    whether to include the size of a file (as "size" column). 
    
  stat : bool
    whether to include the stat of a file (as "stat" column). 
  """
  fdf = _pd.DataFrame({'f': fs})
  if bf:
    fdf = _pdu.update_column(fdf, 'bf', fdf['f'].apply(os.path.basename))
  if bfne:
    fdf = _pdu.update_column(fdf, 'bfne', fdf['f'].apply(os.path.basename)
                                        .apply(remove_complete_ext))
  if size: 
    fdf = _pdu.update_column(fdf, 'size', fdf['f'].apply(file_size))
  if stat: 
    fdf = _pdu.update_column(fdf, 'stat', fdf['f'].apply(os.stat))
  return fdf
  
  
#%% Symlinks
def rel_symlink(d, src, dst):
  """
  Creates a relative symlink, starting at a given directory.
  
  Parameters
  ----------
  d : str
    Directory to temporarily change the current directory to. 
  
  src : str
    Name of the source file.
    
  dst : str
    Name of the destination file. 
  """
  cwd = os.getcwd()
  try:  
    os.chdir(d)
    os.symlink(src, dst)
  finally:
    os.chdir(cwd)
  

def random_hash(n=8):
  """
  Generates `n` random bytes and represents them in a hexadecimal format. 
  
  Parameters
  ----------
  n : int
    Number of bytes to generate.
    
  Examples
  --------
  Let us generate a random hash from 8 bytes:  

  >>> h = random_hash(n=8)
  >>> len(h) # 16: two hex letters for a byte
  16
  """
  return _binascii.b2a_hex(os.urandom(n))


#%% tests
if __name__ == '__main__':
  import doctest
  doctest.testmod()



# -*- coding: utf-8 -*-
# Copyright (c) 2023 Miroslav Hruska
"""
***************
NumPy utilities
***************

Dealing with NumPy arrays. 

Overview
********

Selectors
=========

.. autosummary::
   first_nrows
   each_nth

Run-length encoding
===================

.. autosummary::
   rle

Normalization
=============

.. autosummary::
   onesum_norm

Rounding
========

.. autosummary::
   floor_decimals

Length groups
=============

.. autosummary::
   lengroup_select
   lengroup_starts_ends

Array construction
==================

.. autosummary::
   nonzeros_at

Cumulative sums
===============

.. autosummary::
   zero_cumsum_nolast

Zero/Empty-aware operations
===========================

.. autosummary::
   zaw_1d_concatenate
   eaw_loc

"""

#%% Modules
# global
import numpy as np # skip
import itertools as _it
import fplib as _fplib

#%% * Selectors
def first_nrows(arr, n):
  """
  Selects first `n` rows of a 2-dimensional array `arr`.

  Examples
  --------
  >>> arr = np.array([[1, 2], [2, 3], [3, 4]])
  >>> arr  
  array([[1, 2],
         [2, 3],
         [3, 4]])
  >>> first_nrows(arr, 1)
  array([[1, 2]])
  """
  return arr[:n, :] if not _fplib.zerolenp(arr) else arr
  
  
def each_nth(arr, n=1):
  """
  Selects each `n`-th element of an array `arr`.

  >>> list(each_nth(np.arange(10), 3))
  [0, 3, 6, 9]
  """
  return arr[(np.arange(len(arr)) % n) == 0]


#%% * Run-length encoding
def rle(l):
  """
  Run-length encodes a given list `l`.

  Examples
  --------
  >>> l = ['a', 'a', 'a', 'b', 'c', 'c', 'a']
  >>> rle(l)
  [('a', 3), ('b', 1), ('c', 2), ('a', 1)]
  """
  return [(k, sum(1 for _ in g)) for k, g in _it.groupby(l)]


#%% * Normalization
def onesum_norm(x):
  """
  Normalizes an array `x` to sum to one.

  Examples
  --------
  >>> onesum_norm(np.array([1, 3]))
  array([0.25, 0.75])
  """
  return 1.0 * x / np.sum(x)


#%% * Decimal-level rounding
def floor_decimals(x, dec=0):
  """
  Floors a value `x` to a given number `dec` of decimal places.
  
  Examples
  --------
  >>> x = 123.456

  >>> floor_decimals(x, 0)
  123.0
  >>> floor_decimals(x, 1)
  123.4
  >>> floor_decimals(x, 2)
  123.45
  
  Note the behavior for a negative number of decimals.  
  
  >>> floor_decimals(x, -1)
  120.0
  >>> floor_decimals(x, -2)
  100.0
  >>> floor_decimals(x, -3)
  0.0
  """
  mul = np.power(10.0, dec)
  return np.floor(x * mul) / mul
  

#%% * Length-groups
def lengroup_select(arrs, inds):
  """
  Selects elements from a list of arrays 
  using absolute indices as if the arrays were concatenated.

  Parameters
  ----------
  arrs : list of arrays
    The arrays from which to select elements.

  inds : array_like
    Indices to choose.

  Examples
  --------
  >>> arrs = [np.array([10, 11, 12]), np.array([21, 22])]
  >>> inds = [0, 3]  
  >>> lengroup_select(arrs, inds)
  [10, 21]
  
  >>> arrs_2d = [np.array([[0, 1], [2, 3]]), np.array([[4, 5, 6], [7, 8, 9]])]
  >>> lengroup_select(arrs_2d, inds)
  [array([0, 1]), array([7, 8, 9])]

  >>> lengroup_select(arrs_2d, [])
  []
  """
  clengs = np.array(zero_cumsum_nolast(map(len, arrs)), dtype=np.int)
  # outer array indexes
  out_locs = np.searchsorted(clengs, inds, side='right') - 1
  # inner array indexes
  in_locs = inds - clengs[out_locs]
  return map(lambda o, i: arrs[o][i], out_locs, in_locs)
  
  
def lengroup_outer_inner_indices(lens, lpos):
  """
  Returns the outer and inner indices given lengths `lens` of 
  length-grouped arrays and the desired position `lpos`. 
  
  Examples
  --------
  Suppose we have a length-group array with 2, 4, and 3 elements. 
  
  >>> lens = np.array([2, 4, 3])

  For convenience, let us label the function with a shorter name. 
  
  >>> f = lengroup_outer_inner_indices

  Now, let us obtain all possible absolute indices.   
    
  >>> lpos = np.arange(sum(lens))
  >>> o, i = f(lens, lpos)
  
  Let us take a look at outer indices:   
  >>> list(o)
  [0, 0, 1, 1, 1, 1, 2, 2, 2]
  
  We see that we are getting the proper outer index. 
  
  Let us take a look at inner within-group indices:   
  >>> list(i)
  [0, 1, 0, 1, 2, 3, 0, 1, 2]
  
  Note that an exception is raised if we have out of scope indices.
  
  >>> f(lens, -1)
  Traceback (most recent call last):
  ...
  IndexError: The absolute index is not within limits.

  >>> f(lens, sum(lens))
  Traceback (most recent call last):
  ...
  IndexError: The absolute index is not within limits.

  If we provide empty indices, we also get empty ones. 
  >>> o, i = f(lens, [])
  >>> len(o) == 0 and len(i) == 0
  True
  """
  lpos = np.array(lpos)
  if lpos.size == 0:
    return (np.empty(0, dtype=np.int), np.empty(0, dtype=np.int))
  if np.min(lpos) < 0 or np.max(lpos) >= np.sum(lens):
    raise IndexError("The absolute index is not within limits.")
  clengs = zero_cumsum_nolast(lens)
  oI = np.searchsorted(clengs, lpos, side='right') - 1
  iI = lpos - clengs[oI]
  return oI, iI


def lengroup_starts_ends(lens):
  """
  Obtains starting and ending positions of individual length-groups with 
  lengths specified by `lens`. 
  
  Examples
  --------
  >>> lens = [2, 4, 3]
  >>> s, e = lengroup_starts_ends(lens)  
  >>> s 
  array([0, 2, 6])
  >>> e
  array([2, 6, 9])
  """
  return zero_cumsum_nolast(lens), np.cumsum(lens)
  
  
#%% * Array construction
def nonzeros_at(l, pos, val, dtype):
  """
  Creates an array of a particular length (`l`), 
  by specifying all non-zero values (`val`)
  and their positions (`pos`).
  
  Parameters
  ----------
  dtype : dtype
    The dtype of the array. 
  
  Examples
  --------
  >>> nonzeros_at(l=5, pos=[2, 3], val=1, dtype=np.int)
  array([0, 0, 1, 1, 0])
  """
  T = np.zeros(l, dtype=dtype)
  if len(pos) > 0:
    T[pos] = val
  return T


#%% * Cumulative sums
def zero_cumsum_nolast(arr):
  """
  Returns the cumulative sum of an array `arr`, 
  but starting at the zero and without the last element.
  
  .. note::
    The function is useful when dealing with a list of arrays 
    (of varying lengths). Applying the function over a list of lengths 
    gives starting positions in absolute indices. :func:`~numpy.cumsum` then gives
    the ending positions.
  
  Examples
  --------
  For the sake of illustration, let us show a practical example 
  using a list of strings. 

  >>> seqs = ['abc', 'defgh', 'ij', 'klmnop']  
  
  Let us obtain their lengths:  
  
  >>> arr = map(len, seqs)

  Now we get their starting and ending positions:
  
  >>> starts = zero_cumsum_nolast(arr)
  >>> ends = np.cumsum(arr)  
  
  Suppose the string is concatenated: 
  
  >>> useq = "".join(seqs)  
  >>> useq  
  'abcdefghijklmnop'
  
  We can select the corresponding elements like this:

  >>> i = 1
  >>> s, e = starts[i], ends[i]
  >>> useq[s:e]
  'defgh'
  
  """  
  arr = np.array(arr)
  return np.cumsum(arr) - arr  
  

#%% * Zero/Empty aware functionality
def zaw_1d_concatenate(arrs, default=np.array([])):
  """
  Concatenates 1-d arrays `arrs`, or returns `default` if an empty 
  list was provided.

  Examples
  --------
  >>> default = np.array([], dtype=np.int32)
  >>> zaw_1d_concatenate([], default=default)
  array([], dtype=int32)

  >>> zaw_1d_concatenate([[1, 2], [3]])
  array([1, 2, 3])
  """
  if len(arrs) == 0:
    return default
  else:
    return np.concatenate(arrs)
  
  
def eaw_loc(arr, loc):
  """
  Returns the elements of an array `arr` specified by indices `loc`, 
  and returns an empty array if the indices are empty. 
  
  Examples
  --------
  >>> list(eaw_loc(np.arange(0, 10), []))
  []
  >>> list(eaw_loc([10, 11, 12], [1, 2]))
  [11, 12]
  """
  arr = np.array(arr)
  loc = np.array(loc)  
  if len(loc) == 0:
    return arr[:0]
  else:
    return arr[loc]
  
  
#%% Main
if __name__ == "__main__":
  import doctest
  doctest.testmod()



# -*- coding: utf-8 -*-
# Copyright (c) 2023 Miroslav Hruska
"""
The setup script for the daul package. 
"""

#%% Modules
import os
from setuptools import setup, find_packages

#%% Reading of readme
cwd = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(cwd, 'README.rst')) as h:
  long_description = h.read()

#%% Actual setup
setup(
  name="daul",
  version='0.1.0',
  description="Data analysis and other utilities for python.", 
  long_description=long_description, 
  long_description_content_type='text/x-rst',
  #   
  author="Miroslav Hruska",
  author_email="hruska.miro@gmail.com", 
  #   
  url='https://github.com/hruskamiro/daul', 
  keywords="data analysis utilities",
  classifiers=[
    "License :: OSI Approved :: MIT License",
    "Intended Audience :: Developers",
    "Operating System :: OS Independent", 
  ],
  # 
  packages=find_packages(),
  #   
  python_requires='>=2.7,<3', 
  #   
  install_requires=['numpy>=1.16.5,<1.17', 
                    'pandas>=0.24.2,<0.25', 
                    'fplib>=0.1.1,<0.2',
                    'pathlib>=1.0.1,<1.1']
)
